
// C++ include.
#include <list>
#include <mutex>
#include <memory>
#include <algorithm>
#include <iostream>
#include <cstring>

// boost include.
#include <boost/asio.hpp>
#include <boost/bind.hpp>


using namespace boost::asio::ip;


class Server;


static const int msgSize = 12;

static const char msg[ msgSize ] = { 0x01u, 0x02u, 0x03u, 0x04u, 0x05u, 0x06u,
	0x07u, 0x08u, 0x00u, 0x00u, 0x00u, 0x00u };


//
// Socket
//

class Socket {
public:
	Socket( boost::asio::io_service & service,
		Server * server )
		:	m_socket( service )
		,	m_server( server )
	{
	}

	~Socket()
	{
		std::cout << "Socket destroyed" << std::endl;
	}

	//! \return Socket.
	tcp::socket & socket()
	{
		return m_socket;
	}

	//! Start reading from the socket.
	void start(
		//! Must be this socket.
		std::shared_ptr< Socket > socket );

	//! Stop reading from the socket.
	void stop(
		//! Must be this socket.
		std::shared_ptr< Socket > socket )
	{
		std::cout << "Socket::stop" << std::endl;

		m_socket.close();
	}

private:
	//! Socket.
	tcp::socket m_socket;
	//! Server.
	Server * m_server;
}; // class Socket


//
// Server
//

class Server {
public:
	Server( boost::asio::io_service & service,
		const tcp::endpoint & endpoint )
		:	m_acceptor( service, endpoint )
	{
		startAcceptNewConnection();
	}

	//! Disconnection occured.
	void disconnection( std::shared_ptr< Socket > socket )
	{
		std::cout << "Server::disconnection" << std::endl;

		socket->stop( socket );

		std::lock_guard< std::mutex > lock( m_mutex );

		auto it = std::find( m_clients.begin(), m_clients.end(),
			socket );

		if( it != m_clients.end() )
			m_clients.erase( it );
	}

private:
	//! Start accepting new connection.
	void startAcceptNewConnection()
	{
		std::cout << "Server::startAcceptNewConnection" << std::endl;

		auto socket = std::shared_ptr< Socket >
			( new Socket( m_acceptor.get_io_service(), this ) );

		m_acceptor.async_accept( socket->socket(),
			boost::bind( &Server::handleAccept, this, socket,
				boost::asio::placeholders::error ) );
	}

	//! Handle accept.
	void handleAccept( std::shared_ptr< Socket > socket,
		const boost::system::error_code & error )
	{
		std::cout << "Server::handleAccept" << std::endl;

		if( !error )
		{
			std::lock_guard< std::mutex > lock( m_mutex );

			socket->start( socket );

			m_clients.push_back( socket );
		}

		startAcceptNewConnection();
	}

private:
	//! Acceptor.
	tcp::acceptor m_acceptor;
	//! List of sockets.
	std::list< std::shared_ptr< Socket > > m_clients;
	//! Mutex.
	std::mutex m_mutex;
}; // class Server


void
Socket::start( std::shared_ptr< Socket > socket )
{
	std::cout << "Socket::start" << std::endl;

	char buf[ msgSize ];

	boost::asio::async_read( m_socket,
		boost::asio::buffer( buf, msgSize ),
		[ this, socket, &buf ] ( boost::system::error_code error, std::size_t )
			{
				if( error )
					m_server->disconnection( socket );
				else
				{
					if( std::strncmp( buf, msg, msgSize ) != 0 )
					{
						std::cout << "Message is WRONG" << std::endl;

						m_server->disconnection( socket );
					}
					else if( m_socket.is_open() )
					{
						std::cout << "Message is OK" << std::endl;

						start( socket );
					}
				}
			}
	);
}


int main( int, char ** )
{
	boost::asio::io_service service;
	Server server( service, tcp::endpoint( tcp::v4(), 4545 ) );

	service.run();

	return 0;
}
